class Minsky:
    """Simulateur de machines à compteurs."""

    def __init__(self,instructions,état_initial):
        self.instructions = instructions
        # utilisé uniquement pour l’affichage avec __str__() ci-dessous
        self._largeur  = max(*(len(état) for état in instructions))
        self.état      = état_initial

    def calcule_un_pas(self):
        """Effectue une étape de la machine suivant son instructions."""
        instruction = self.instructions[self.état]
        if len(instruction) == 2:
            id_compteur, self.état = instruction
            self.compteurs[id_compteur] += 1
        else:
            id_compteur, état_1, état_0 = instruction
            compteur = self.compteurs[id_compteur]
            if compteur > 0:
                self.compteurs[id_compteur] = compteur - 1
                self.état = état_1
            else:
                self.état = état_0
    
    def calcule(self, compteurs, trace=False):
        """Fait tourner la machine: le dictionnaire `compteurs`
        donne les valeurs initiales.
        
        Si `trace` est mis à `True`, on affiche les étapes successives."""

        self.compteurs = compteurs

        if trace: print(self)
        while self.état in self.instructions:
            self.calcule_un_pas()
            if trace: print(self)

    def __str__(self):
        return f"{self.état:{self._largeur}}: {self.compteurs}"

if __name__ == "__main__":
    compteurs = [3,4,0,0]
    additionne_0_et_1_vers_2_via_3 = {
            "0":    (0, "0>2", "3>0"),
            "0>2":  (2, "0>3"),
            "0>3":  (3, "0"),
            "3>0":  (3, "3>0+", "1"),
            "3>0+": (0, "3>0"),
            "1":    (1, "1>2", "3>1"),
            "1>2":  (2, "1>3"),
            "1>3":  (3, "1"),
            "3>1":  (3, "3>1+", "F"),
            "3>1+": (1, "3>1"),
    }
    Minsky(additionne_0_et_1_vers_2_via_3,"0").calcule(compteurs,trace=True)
